from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarSectionLowercase(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.sectionlowercase',
                         severity='warning',
                         message='\'SECTION\' should only lowercase characters')

    def __getMatches(self, _file: str, stash: Stash) -> List[Variable]:
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue='SECTION')
        for i in items:
            if not i.VarValue.islower():
                res.append(i)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = i.RealRaw.replace(i.VarValue, i.VarValue.lower())
            i.Raw = i.Raw.replace(i.VarValue, i.VarValue.lower())
            i.VarValue = i.VarValue.lower()
            res.append(_file)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarDescSameTooBrief(Rule):
    def __init__(self):
        super().__init__(id='oelint.vars.descriptiontoobrief',
                         severity='warning',
                         message='\'DESCRIPTION\' is the shorter than \'SUMMARY\'')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='DESCRIPTION')
        items_sum: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                      attribute=Variable.ATTR_VAR, attributeValue='SUMMARY')
        for i in items:
            _same = [x for x in items_sum if len(x.VarValueStripped) > len(i.VarValueStripped)]
            if any(_same):
                res += self.finding(i.Origin, i.InFileLine)
        return res



import os
from typing import List, Tuple

import regex  # noqa: I900
from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule


class FilePatchIsUpstreamStatus(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.file.upstreamstatus',
                         severity='info',
                         message='Patch \'{FILE}\' should contain an Upstream-Status entry')

    def _get_recipe(self, items, path):
        # Find matching SRC_URI assignment
        return [x for x in items if x.VarValue.find(os.path.basename(path)) != -1]

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        patches = stash.GetFiles(_file, '*.patch')
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue='SRC_URI')

        _valid_class = {
            'Pending': r'Pending',
            'Submitted': r'Submitted(\s+\[.*\])*',
            'Accepted': r'Accepted',
            'Denied': r'Denied',
            'Backport': r'Backport',
            'Inappropriate': r'Inappropriate(\s+\[.*\])*',
            'Inactive-Upstream': r'Inactive-Upstream(\s+\[.*\])*',
        }
        for i in patches:
            with open(i) as _input:
                found = False
                _recipe_match = self._get_recipe(_items, i)
                if _recipe_match:
                    _recipe_match = _recipe_match[0]
                else:
                    continue  # pragma: no cover
                try:
                    cnt = _input.read()
                    for m in RegexRpl.finditer(r'^Upstream-Status:\s*(?P<class>.*)', cnt, flags=regex.regex.MULTILINE):
                        found = True
                        if not any(RegexRpl.match(v, m.group('class')) for k, v in _valid_class.items()):
                            _msg = 'Upstream-Status in \'{FILE}\' doesn\'t pick from valid classifiers {cls}'.format(
                                FILE=os.path.basename(i), cls=','.join(sorted(_valid_class.keys())),
                            )
                            res += self.finding(_recipe_match.Origin,
                                                _recipe_match.InFileLine,
                                                override_msg=_msg)
                    if not found:
                        res += self.finding(_recipe_match.Origin,
                                            _recipe_match.InFileLine,
                                            self.Msg.replace('{FILE}', os.path.basename(i)))
                except UnicodeDecodeError:  # pragma: no cover
                    pass  # pragma: no cover
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Function, Item, PythonBlock
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class NoSpaceBeginningRule(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.spaces.linebeginning',
                         severity='warning',
                         message='Line shall not begin with a space')

    def __getMatches(self, _file: str, stash: Stash) -> List[Item]:
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if i.Raw and i.Raw.startswith(' '):
                res.append(i)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = []
        for i in self.__getMatches(_file, stash):
            if isinstance(i, PythonBlock) or isinstance(i, Function):
                continue  # pragma: no cover
            i.RealRaw = i.RealRaw.lstrip(' ')
            i.Raw = i.Raw.lstrip(' ')
            res.append(_file)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Inherit
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarMultiInherit(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.multiinherit',
                         severity='warning',
                         message='\'{inherit}\' is included multiple times')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Inherit] = stash.GetItemsFor(filename=_file, classifier=Inherit.CLASSIFIER)
        keys = []
        for i in items:
            for y in i.get_items():
                if y not in keys:
                    keys.append(y)
                else:
                    res += self.finding(i.Origin, i.InFileLine,
                                        self.Msg.replace('{inherit}', y))
        return res



import os
from typing import List, Tuple

from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class BBClassUnderscore(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.bbclass.underscores',
                         severity='error',
                         message='bbclass filenames should not contain \'_\'. Replace it by \'-\'')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _files = {x.Origin for x in stash.GetItemsFor(filename=_file) if x.Origin.endswith('.bbclass')}
        for file in _files:
            if '_' in os.path.basename(file):
                res += self.finding(file, 1)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarOverrideAppend(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.overrideappend',
                         severity='warning',
                         message='This creates an empty scope which is then appended. It should be {op}:{_class} instead')
        self.pkgspecific = [
            'ALLOW_EMPTY',
            'ALTERNATIVE',
            'CONFFILES',
            'DEBIAN_NOAUTONAME',
            'DEBIANNAME',
            'DEPENDS',
            'DESCRIPTION',
            'FILES',
            'GROUPADD_PARAM',
            'INITSCRIPT_NAME',
            'INITSCRIPT_PARAMS',
            'INSANE_SKIP',
            'LICENSE',
            'PKG',
            'pkg_postinst',
            'pkg_postinst_ontarget',
            'pkg_postrm',
            'pkg_preinst',
            'pkg_prerm',
            'PRIVATE_LIBS',
            'RCONFLICTS',
            'RDEPENDS',
            'RPROVIDES',
            'RRECOMMENDS',
            'RREPLACES',
            'RSUGGESTS',
            'SECTION',
            'SKIP_FILEDEPS',
            'sstate',
            'SUMMARY',
            'SYSTEMD_AUTO_ENABLE',
            'SYSTEMD_SERVICE',
            'USERADD_PARAM',
        ]

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR)

        _store = set()

        for i in sorted(items, key=lambda key: key.Line):
            _items = [x for x in i.SubItems if x]
            if i.VarName in self.pkgspecific:
                _items = _items[1:]
            _scope = [x for x in _items if x not in ['append', 'prepend']]
            _op = [x for x in _items if x in ['append', 'prepend']]
            if not _scope or not _items:
                continue  # pragma: no cover
            _scope_key = i.VarName + ':' + '-'.join(_scope)
            if _op:
                if _items[0] not in _op and _scope_key not in _store:
                    res += self.finding(i.Origin, i.InFileLine,
                                        override_msg=self.Msg.format(op=_op[0], _class=':'.join(_scope)))
            _store.add(_scope_key)
        return res



from typing import List, Tuple
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarHomepagePing(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.homepageping',
                         severity='warning',
                         message='\'HOMEPAGE\' isn\'t reachable')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='HOMEPAGE')
        for i in items:
            try:
                req = Request(i.VarValueStripped)  # noqa: S310 - we can take the risk of calling unexpected schemes here
                try:
                    urlopen(req, timeout=4)  # noqa: S310 - we can take the risk of calling unexpected schemes here
                except HTTPError as e:
                    if e.code == 404:  # pragma: no cover
                        res += self.finding(i.Origin, i.InFileLine)  # pragma: no cover
                except URLError:
                    res += self.finding(i.Origin, i.InFileLine)
            except ValueError:
                res += self.finding(i.Origin, i.InFileLine)
            except Exception:  # noqa: S110, pragma: no cover
                pass  # pragma: no cover
        return res



from collections import OrderedDict
from typing import List, Tuple

from oelint_parser.cls_item import (
    Export,
    Function,
    FunctionExports,
    Item,
    PythonBlock,
    Variable,
)
from oelint_parser.cls_stash import Stash
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule


class VarsPathHardcode(Rule):
    def __init__(self) -> None:
        self._map = OrderedDict({
            '/usr/lib/systemd/user': '${systemd_user_unitdir}',
            '/lib/systemd/system': '${systemd_system_unitdir}',
            '/usr/share/doc': '${docdir}',
            '/usr/share/info': '${infodir}',
            '/usr/share/man': '${mandir}',
            '/usr/libexec': '${libexecdir}',
            '/lib/systemd': '${systemd_unitdir}',
            '/usr/lib': '${libdir}',
            '/usr/bin': '${bindir}',
            '/usr/share': '${datadir}',
            '/usr/include': '${includedir}',
            '/var': '${localstatedir}',
            '/lib': '${nonarch_base_libdir}',
            '/usr/sbin': '${sbindir}',
            '/srv': '${servicedir}',
            '/com': '${sharedstatedir}',
            '/etc': '${sysconfdir}',
        })
        super().__init__(id='oelint.vars.pathhardcode',
                         severity='warning',
                         message='<FOO>',
                         appendix=[v.strip('$').strip('{').strip('}') for v in self._map.values()])

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Item] = stash.GetItemsFor(filename=_file,
                                              classifier=[Variable.CLASSIFIER, Export.CLASSIFIER,
                                                          Function.CLASSIFIER, FunctionExports.CLASSIFIER,
                                                          PythonBlock.CLASSIFIER])

        for i in items:
            if isinstance(i, Variable) and i.VarName in ['SUMMARY', 'DESCRIPTION', 'HOMEPAGE', 'AUTHOR',
                                                         'BUGTRACKER', 'FILES', 'CVE_STATUS']:
                continue
            _matches = []
            for k, v in self._map.items():
                for pre in ['^', '\'', '"', ' ', r'\$\{D\}', '=']:
                    for ext in [' ', '/', '"', r'\*', '$']:
                        for line in i.get_items():
                            if line.strip().startswith('#'):
                                continue
                            _match = RegexRpl.search(pre + k + ext, line)
                            if _match and k not in _matches:
                                _cleanapp = v.strip('$').strip('{').strip('}')
                                res += self.finding(i.Origin, i.InFileLine,
                                                    '\'{a}\' should be used instead of \'{b}\''.format(a=v, b=k), appendix=_cleanapp)
                                _parts = [x for x in k.split('/') if x]
                                for x in range(1, len(_parts) + 1):
                                    _k = '/' + '/'.join(_parts[0:x])
                                    _matches.append(_k)
        return res
